#!/usr/bin/python
# Copyright 2009 Canonical Ltd.
#
# This file is part of desktopcouch.
#
#  desktopcouch is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation.
#
# desktopcouch is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with desktopcouch.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Stuart Langridge <stuart.langridge@canonical.com>
"""
Start local CouchDB server.
Steps:
    1. Work out which folders to use (running from source tree, or installed)
    2. Actually start CouchDB
    3. check design documents in central folder against design documents in
       Couch, and overwrite any new ones
    4. Write an HTML "bookmark" file which directs people to the local
       running CouchDB.

This script is normally called by advertisePort.py, which advertises the local
CouchDB port over D-Bus. That advertisePort script is started by D-Bus
activation.
"""

from __future__ import with_statement
import os, subprocess, sys
import desktopcouch
from desktopcouch import local_files
import xdg.BaseDirectory
import time

def dump_ini(data, filename):
    """Dump INI data with sorted sections and keywords"""
    fd = open(filename, 'w')
    sections = data.keys()
    sections.sort()
    for section in sections:
        fd.write("[%s]\n" % (section))
        keys = data[section].keys()
        keys.sort()
        for key in keys:
            fd.write("%s=%s\n" % (key, data[section][key]))
        fd.write("\n")
    fd.close()



def create_ini_file():
    """Write CouchDB ini file if not already present"""
    # FIXME add update trigger folder
    #update_trigger_dir = [
    #    'lib', 'canonical', 'ubuntuone', 'cloud_server', 'update_triggers']
    #
    #timestamp_trigger = os.path.join(
    #    *update_trigger_dir + ['timestamp_trigger.py'])
    #update_trigger =  os.path.join(
    #    *update_trigger_dir + ['update_trigger.py'])

    if os.path.exists(local_files.FILE_INI):
        return

    local = {
        'couchdb': {
            'database_dir': local_files.DIR_DB,
            'view_index_dir': local_files.DIR_DB,
        },
        'httpd': {
            'bind_address': '127.0.0.1',
            'port': "0",
        },
        'log': {
            'file': local_files.FILE_LOG,
            'level': 'info',
        },
    }

    dump_ini(local, local_files.FILE_INI)

def run_couchdb():
    """Actually start the CouchDB process"""
    local_exec = local_files.COUCH_EXEC_COMMAND + ['-b']
    try:
        retcode = subprocess.call(local_exec, shell=False)
        if retcode < 0:
            print >> sys.stderr, "Child was terminated by signal", -retcode
        elif retcode > 0:
            print >> sys.stderr, "Child returned", retcode
    except OSError, e:
        print >> sys.stderr, "Execution failed: %s: %s" % (e, local_exec)
        exit(1)

def update_design_documents():
    """Check system design documents and update any that need updating"""
    pass

def write_bookmark_file():
    """Write out an HTML document that the user can bookmark to find their DB"""
    bookmark_file = os.path.join(local_files.DIR_DB, "couchdb.html")

    if os.path.exists(os.path.join(os.path.split(__file__)[0], "../data/couchdb.tmpl")):
        bookmark_template = os.path.join(os.path.split(__file__)[0], "../data/couchdb.tmpl")
    else:
        for base in xdg.BaseDirectory.xdg_data_dirs:
            template_path = os.path.join(base, "desktopcouch", "couchdb.tmpl")
            if os.path.exists(template_path):
                bookmark_template = os.path.join(os.path.split(__file__)[0], template_path)

    fp = open(bookmark_template)
    html = fp.read()
    fp.close()

    time.sleep(1)
    pid = desktopcouch.find_pid()
    port = desktopcouch.find_port(pid)

    fp = open(bookmark_file, "w")
    fp.write(html.replace("[[COUCHDB_PORT]]", port))
    fp.close()
    print "Browse your desktop CouchDB at file://%s" % \
      os.path.realpath(bookmark_file)

def start_couchdb():
    """Execute each step to start a desktop CouchDB"""
    create_ini_file()
    run_couchdb()
    update_design_documents()
    write_bookmark_file()

if __name__ == "__main__":
    start_couchdb()
    print "Desktop CouchDB started"

